import mock

import uuid

from nsq import reader
from nsq import response

from common import HttpClientIntegrationTest


class TestReader(HttpClientIntegrationTest):
    '''Tests for our reader class'''
    nsqd_ports = (14150, 14152)

    def setUp(self):
        '''Return a connection'''
        HttpClientIntegrationTest.setUp(self)
        nsqd_tcp_addresses = ['localhost:%s' % port for port in self.nsqd_ports]
        self.client = reader.Reader(
            self.topic, self.channel, nsqd_tcp_addresses=nsqd_tcp_addresses)

    def test_it_subscribes(self):
        '''It subscribes for newly-established connections'''
        connection = mock.Mock()
        self.client.added(connection)
        connection.sub.assert_called_with(self.topic, self.channel)

    def test_new_connections_rdy(self):
        '''Calls rdy(1) when connections are added'''
        connection = mock.Mock()
        self.client.added(connection)
        connection.rdy.assert_called_with(1)

    def test_reconnected_rdy(self):
        '''Calls rdy(1) when connections are reestablished'''
        connection = mock.Mock()
        self.client.reconnected(connection)
        connection.rdy.assert_called_with(1)

    def test_added_dead(self):
        '''Does not call reconnected when adding dead connections'''
        conn = mock.Mock()
        conn.alive.return_value = False
        with mock.patch.object(self.client, 'reconnected') as mock_reconnected:
            self.client.added(conn)
            self.assertFalse(mock_reconnected.called)

    def test_it_checks_max_in_flight(self):
        '''Raises an exception if more connections than in-flight limit'''
        with mock.patch.object(self.client, '_max_in_flight', 0):
            self.assertRaises(NotImplementedError, self.client.distribute_ready)

    def test_it_distributes_ready(self):
        '''It distributes RDY with util.distribute'''
        with mock.patch('nsq.reader.distribute') as mock_distribute:
            counts = range(10)
            connections = [mock.Mock() for _ in counts]
            mock_distribute.return_value = zip(counts, connections)
            self.client.distribute_ready()
            for count, connection in zip(counts, connections):
                connection.rdy.assert_called_with(count)

    def test_it_ignores_dead_connections(self):
        '''It does not distribute RDY state to dead connections'''
        dead = mock.Mock()
        dead.alive.return_value = False
        alive = mock.Mock()
        alive.alive.return_value = True
        with mock.patch.object(
            self.client, 'connections', return_value=[alive, dead]):
            self.client.distribute_ready()
            self.assertTrue(alive.rdy.called)
            self.assertFalse(dead.rdy.called)

    def test_zero_ready(self):
        '''When a connection has ready=0, distribute_ready is invoked'''
        connection = self.client.connections()[0]
        with mock.patch.object(connection, 'ready', 0):
            self.assertTrue(self.client.needs_distribute_ready())

    def test_not_ready(self):
        '''When no connection has ready=0, distribute_ready is not invoked'''
        connection = self.client.connections()[0]
        with mock.patch.object(connection, 'ready', 10):
            self.assertFalse(self.client.needs_distribute_ready())

    def test_negative_ready(self):
        '''If clients have negative RDY values, distribute_ready is invoked'''
        connection = self.client.connections()[0]
        with mock.patch.object(connection, 'ready', -1):
            self.assertTrue(self.client.needs_distribute_ready())

    def test_low_ready(self):
        '''If clients have negative RDY values, distribute_ready is invoked'''
        connection = self.client.connections()[0]
        with mock.patch.object(connection, 'ready', 2):
            with mock.patch.object(connection, 'last_ready_sent', 10):
                self.assertTrue(self.client.needs_distribute_ready())

    def test_none_alive(self):
        '''We don't need to redistribute RDY if there are none alive'''
        with mock.patch.object(self.client, 'connections', return_value=[]):
            self.assertFalse(self.client.needs_distribute_ready())

    def test_read_distribute_ready(self):
        '''Read checks if we need to distribute ready'''
        with mock.patch('nsq.reader.Client'):
            with mock.patch.object(
                self.client, 'needs_distribute_ready', return_value=True):
                with mock.patch.object(
                    self.client, 'distribute_ready') as mock_ready:
                    self.client.read()
                    mock_ready.assert_called_with()

    def test_read_not_distribute_ready(self):
        '''Does not redistribute ready if not needed'''
        with mock.patch('nsq.reader.Client'):
            with mock.patch.object(
                self.client, 'needs_distribute_ready', return_value=False):
                with mock.patch.object(
                    self.client, 'distribute_ready') as mock_ready:
                    self.client.read()
                    self.assertFalse(mock_ready.called)

    def test_iter(self):
        '''The client can be used as an iterator'''
        iterator = iter(self.client)
        message_id = uuid.uuid4().hex[0:16]
        packed = response.Message.pack(0, 0, message_id, 'hello')
        messages = [response.Message(None, None, packed) for _ in range(10)]
        with mock.patch.object(self.client, 'read', return_value=messages):
            found = [iterator.next() for _ in range(10)]
            self.assertEqual(messages, found)

    def test_iter_repeated_read(self):
        '''Repeatedly calls read in iterator mode'''
        iterator = iter(self.client)
        message_id = uuid.uuid4().hex[0:16]
        packed = response.Message.pack(0, 0, message_id, 'hello')
        messages = [response.Message(None, None, packed) for _ in range(10)]
        for message in messages:
            with mock.patch.object(self.client, 'read', return_value=[message]):
                self.assertEqual(iterator.next(), message)

    def test_skip_non_messages(self):
        '''Skips all non-messages'''
        iterator = iter(self.client)
        message_id = uuid.uuid4().hex[0:16]
        packed = response.Message.pack(0, 0, message_id, 'hello')
        messages = [response.Message(None, None, packed) for _ in range(10)]
        packed = response.Response.pack('hello')
        responses = [
            response.Response(None, None, packed) for _ in range(10)] + messages
        with mock.patch.object(self.client, 'read', return_value=responses):
            found = [iterator.next() for _ in range(10)]
            self.assertEqual(messages, found)

    def test_honors_max_rdy_count(self):
        '''Honors the max RDY count provided in an identify response'''
        for conn in self.client.connections():
            conn.max_rdy_count = 10
        self.client.distribute_ready()
        self.assertEqual(self.client.connections()[0].ready, 10)

    def test_read(self):
        '''Can receive a message in a basic way'''
        self.nsqd.pub(self.topic, 'hello')
        message = iter(self.client).next()
        self.assertEqual(message.body, 'hello')

    def test_close_redistribute(self):
        '''Redistributes rdy count when a connection is closed'''
        with mock.patch.object(self.client, 'distribute_ready') as mock_ready:
            self.client.close_connection(self.client.connections()[0])
            mock_ready.assert_called_with()



'''Test our stats utility'''

import mock
import os
import unittest

import simplejson as json

from nsq.stats import Nsqlookupd
from nsq.http import nsqd
from common import IntegrationTest


class TestStats(IntegrationTest):
    '''Test our stats utility.'''

    nsqd_ports = (14150, 14152)

    def assertFixture(self, path, actual):
        '''Assert actual is equivalent to the JSON fixture provided.'''
        if os.environ.get('RECORD', 'false').lower() == 'true':
            with open(path, 'w') as fout:
                json.dump(actual,
                    fout, sort_keys=True, indent=4, separators=(',', ': ' ))
                # Add a trailing newline
                fout.write('\n')
        else:
            with open(path) as fin:
                self.assertEqual(json.load(fin), actual)

    def test_stats(self):
        '''Can effectively grab all the stats.'''
        # Create some topics, channels, and messages on the first nsqd instance
        client = nsqd.Client('http://localhost:14151')
        topics = [
            'topic-on-both-instances',
            'topic-with-channels',
            'topic-without-channels'
        ]
        for topic in topics:
            client.create_topic(topic)
            client.mpub(topic, map(str, range(len(topic))))
        client.create_channel('topic-with-channels', 'channel')

        # Create a topic and messages on the second nsqd isntance
        client = nsqd.Client('http://localhost:14153')
        client.create_topic('topic-on-both-instances')
        client.mpub(topic, map(str, range(10)))

        # Check the stats
        self.assertFixture(
            'test/fixtures/test_stats/TestStats/stats',
            map(list, Nsqlookupd('http://localhost:14161').stats))



'''Wraps a socket in TLS'''

import ssl

from .. import logger


class TLSSocket(object):
    '''Provide a way to return a TLS socket'''
    @classmethod
    def wrap_socket(cls, socket):
        sock = ssl.wrap_socket(socket, ssl_version=ssl.PROTOCOL_TLSv1)
        while True:
            try:
                logger.info('Performing TLS handshade...')
                sock.do_handshake()
                break
            except ssl.SSLError as err:
                errs = (
                    ssl.SSL_ERROR_WANT_READ,
                    ssl.SSL_ERROR_WANT_WRITE)
                if err.args[0] not in (errs):
                    raise
                else:
                    logger.info('Continuing TLS handshake...')
        logger.info('Socket wrapped')
        return sock



from collections import Counter

from .http import nsqlookupd, nsqd

class Nsqlookupd(object):
    '''A class for grabbing stats about all hosts and topics reported to nsqlookupd.'''
    def __init__(self, *args, **kwargs):
        self.client = nsqlookupd.Client(*args, **kwargs)

    @property
    def merged(self):
        '''The clean stats from all the hosts reporting to this host.'''
        stats = {}
        for topic in self.client.topics()['topics']:
            for producer in self.client.lookup(topic)['producers']:
                hostname = producer['broadcast_address']
                port = producer['http_port']
                host = '%s_%s' % (hostname, port)
                stats[host] = nsqd.Client(
                    'http://%s:%s/' % (hostname, port)).clean_stats()
        return stats

    @property
    def raw(self):
        '''All the raw, unaggregated stats (with duplicates).'''
        topic_keys = (
            'message_count',
            'depth',
            'backend_depth',
            'paused'
        )

        channel_keys = (
            'in_flight_count',
            'timeout_count',
            'paused',
            'deferred_count',
            'message_count',
            'depth',
            'backend_depth',
            'requeue_count'
        )
        
        for host, stats in self.merged.items():
            for topic, stats in stats.get('topics', {}).items():
                prefix = 'host.%s.topic.%s' % (host, topic)
                for key in topic_keys:
                    value = int(stats.get(key, -1))
                    yield (
                        'host.%s.topic.%s.%s' % (host, topic, key),
                        value,
                        False
                    )
                    yield (
                        'topic.%s.%s' % (topic, key),
                        value,
                        True
                    )
                    yield (
                        'topics.%s' % key,
                        value,
                        True
                    )
                
                for chan, stats in stats.get('channels', {}).items():
                    data = {
                        key: int(stats.get(key, -1)) for key in channel_keys
                    }
                    data['clients'] = len(stats.get('clients', []))

                    for key, value in data.items():
                        yield (
                            'host.%s.topic.%s.channel.%s.%s' % (host, topic, chan, key),
                            value,
                            False
                        )
                        yield (
                            'host.%s.topic.%s.channels.%s' % (host, topic, key),
                            value,
                            True
                        )
                        yield (
                            'topic.%s.channels.%s' % (topic, key),
                            value,
                            True
                        )
                        yield (
                            'channels.%s' % key,
                            value,
                            True
                        )

    @property
    def stats(self):
        '''Stats that have been aggregated appropriately.'''
        data = Counter()
        for name, value, aggregated in self.raw:
            if aggregated:
                data['%s.max' % name] = max(data['%s.max' % name], value)
                data['%s.total' % name] += value
            else:
                data[name] = value

        return sorted(data.items())



import mock
import unittest

from nsq import backoff


class TestBackoff(unittest.TestCase):
    '''Test our backoff class'''
    def setUp(self):
        self.backoff = backoff.Backoff()

    def test_sleep(self):
        '''Just calls time.sleep with whatever backoff returns'''
        with mock.patch.object(self.backoff, 'backoff', return_value=5):
            with mock.patch('nsq.backoff.time') as MockTime:
                self.backoff.sleep(10)
                self.backoff.backoff.assert_called_with(10)
                MockTime.sleep.assert_called_with(5)

    def test_backoff(self):
        '''Not implemented on the base class'''
        self.assertRaises(NotImplementedError, self.backoff.backoff, 1)


class TestLinear(unittest.TestCase):
    '''Test our linear backoff class'''
    def setUp(self):
        self.backoff = backoff.Linear(1, 2)

    def test_constant(self):
        '''The constant is added to each time'''
        self.assertEqual(self.backoff.backoff(0), 2)

    def test_affine(self):
        '''The affine factor works as advertised'''
        self.assertEqual(self.backoff.backoff(0) + 1, self.backoff.backoff(1))


class TestConstant(unittest.TestCase):
    '''Test our constant backoff class'''
    def setUp(self):
        self.backoff = backoff.Constant(10)

    def test_constant(self):
        '''Always gives the same result'''
        for i in xrange(100):
            self.assertEqual(self.backoff.backoff(0), self.backoff.backoff(i))


class TestExponential(unittest.TestCase):
    '''Test our exponential backoff class'''
    def test_factor(self):
        '''We make use of the constant factor'''
        base = 5
        one = backoff.Exponential(base, 1)
        two = backoff.Exponential(base, 2)
        for i in xrange(10):
            self.assertEqual(one.backoff(i) * 2, two.backoff(i))

    def test_constant(self):
        '''We add the constant value'''
        base = 5
        four = backoff.Exponential(base, c=4)
        zero = backoff.Exponential(base)
        for i in xrange(10):
            self.assertEqual(zero.backoff(i) + 4, four.backoff(i))

    def test_base(self):
        '''We honor the base'''
        one = backoff.Exponential(1)
        two = backoff.Exponential(2)
        self.assertEqual(one.backoff(1) * 2, two.backoff(1))
        self.assertEqual(one.backoff(2) * 4, two.backoff(2))


class TestClamped(unittest.TestCase):
    '''Does in fact keep our backoff clamped'''
    def setUp(self):
        self.linear = backoff.Linear(1, 2)
        self.backoff = backoff.Clamped(self.linear, minimum=5, maximum=10)

    def test_min(self):
        '''Asserts a minimum'''
        self.assertLess(self.linear.backoff(0), 5)
        self.assertEqual(self.backoff.backoff(0), 5)

    def test_max(self):
        '''Asserts a maximum'''
        self.assertGreater(self.linear.backoff(100), 10)
        self.assertEqual(self.backoff.backoff(100), 10)


class TestAttemptCounter(unittest.TestCase):
    '''Test the attempt counter'''
    def setUp(self):
        self.backoff = mock.Mock()
        self.counter = backoff.AttemptCounter(self.backoff)

    def test_sleep(self):
        '''Just calls time.sleep with whatever backoff returns'''
        with mock.patch.object(self.counter, 'backoff', return_value=5):
            with mock.patch('nsq.backoff.time') as MockTime:
                self.counter.sleep()
                self.counter.backoff.assert_called_with()
                MockTime.sleep.assert_called_with(5)

    def test_backoff(self):
        '''Not implemented on the base class'''
        with mock.patch.object(self.counter, 'attempts', 5):
            self.assertEqual(
                self.counter.backoff(), self.backoff.backoff.return_value)
            self.backoff.backoff.assert_called_with(5)

    def test_failed(self):
        '''Failed increments the number of attempts'''
        for attempts in range(10):
            self.assertEqual(self.counter.attempts, attempts)
            self.counter.failed()

    def test_ready_false(self):
        '''Ready returns false if not enough time has elapsed'''
        with mock.patch('nsq.backoff.time') as mock_time:
            mock_time.time = mock.Mock(return_value=10)
            with mock.patch.object(self.counter, '_last_failed', 10):
                self.assertFalse(self.counter.ready())

    def test_ready_true(self):
        '''Ready returns true if enough time has elapsed'''
        with mock.patch('nsq.backoff.time') as mock_time:
            mock_time.time = mock.Mock(return_value=10)
            with mock.patch.object(self.counter, '_last_failed', 1):
                with mock.patch.object(self.counter, 'backoff', return_value=5):
                    self.assertTrue(self.counter.ready())

    def test_ready_never_failed(self):
        '''If it has never failed, then it returns True'''
        with mock.patch.object(self.counter, '_last_failed', None):
            self.assertTrue(self.counter.ready())


class TestResettingAttemptCounter(unittest.TestCase):
    '''Test the ResettingAttemptCounter'''
    def setUp(self):
        self.counter = backoff.ResettingAttemptCounter(None)

    def test_success(self):
        '''Success resets the attempts counter'''
        for _ in range(10):
            self.counter.failed()
        self.counter.success()
        self.assertEqual(self.counter.attempts, 0)


class TestDecrementingAttemptCounter(unittest.TestCase):
    def setUp(self):
        self.counter = backoff.DecrementingAttemptCounter(None)

    def test_success(self):
        '''Success only decrements the attempts counter'''
        for _ in range(10):
            self.counter.failed()
        self.counter.success()
        self.assertEqual(self.counter.attempts, 9)

    def test_negative_attempts(self):
        '''Success never lets the attempts count drop below 0'''
        self.counter.success()
        self.assertEqual(self.counter.attempts, 0)



#! /usr/bin/env python

import mock

import errno
import socket
import ssl
from collections import deque

from nsq import connection
from nsq import constants
from nsq import exceptions
from nsq import response
from nsq import util
from nsq import json
from common import MockedSocketTest, HttpClientIntegrationTest


class TestConnection(MockedSocketTest):
    '''Tests about our connection'''
    def test_alive(self):
        self.assertTrue(self.connection.alive())

    def test_close(self):
        '''Should mark the connection as closed'''
        self.connection.close()
        self.assertFalse(self.connection.alive())

    def test_blocking(self):
        '''Sets blocking on the socket'''
        self.connection.setblocking(0)
        self.socket.setblocking.assert_called_with(0)

    def test_pending(self):
        '''Appends to pending'''
        self.connection.nop()
        self.assertEqual(
            list(self.connection.pending()), [constants.NOP + constants.NL])

    def test_flush_partial(self):
        '''Keeps its place when flushing out partial messages'''
        # We'll tell the connection it has only sent one byte when flushing
        with mock.patch.object(self.socket, 'send'):
            self.socket.send.return_value = 1
            self.connection.nop()
            self.connection.flush()
            # We expect all but the first byte to remain
            message = constants.NOP + constants.NL
            self.assertEqual(list(self.connection.pending()), [message[1:]])

    def test_flush_full(self):
        '''Pops off messages it has flushed completely'''
        # We'll tell the connection it has only sent one byte when flushing
        self.connection.nop()
        self.connection.flush()
        # The nop message was sent, so we expect it to be popped
        self.assertEqual(list(self.connection.pending()), [])

    def test_flush_count(self):
        '''Returns how many bytes were sent'''
        message = constants.NOP + constants.NL
        # Ensure this doesn't invoke our normal flush
        self.connection.nop()
        self.assertEqual(self.connection.flush(), len(message))

    def test_flush_empty(self):
        '''Returns 0 if there are no pending messages'''
        self.assertEqual(self.connection.flush(), 0)

    def test_flush_multiple(self):
        '''Flushes as many messages as possible'''
        pending = deque(['hello'] * 5)
        with mock.patch.object(self.connection, '_pending', pending):
            self.connection.flush()
            self.assertEqual(len(self.connection.pending()), 0)

    def test_flush_would_block(self):
        '''Honors EAGAIN / EWOULDBLOCK'''
        pending = deque(map(str, [1, 2, 3]))
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            with mock.patch.object(self.connection, '_pending', pending):
                mock_socket.send.side_effect = socket.error(errno.EAGAIN)
                self.assertEqual(self.connection.flush(), 0)

    def test_flush_would_block_ssl_write(self):
        '''Honors ssl.SSL_ERROR_WANT_WRITE'''
        pending = deque(map(str, [1, 2, 3]))
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            with mock.patch.object(self.connection, '_pending', pending):
                mock_socket.send.side_effect = ssl.SSLError(
                    ssl.SSL_ERROR_WANT_WRITE)
                self.assertEqual(self.connection.flush(), 0)

    def test_flush_would_block_ssl_read(self):
        '''Honors ssl.SSL_ERROR_WANT_READ'''
        pending = deque(map(str, [1, 2, 3]))
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            with mock.patch.object(self.connection, '_pending', pending):
                mock_socket.send.side_effect = ssl.SSLError(
                    ssl.SSL_ERROR_WANT_READ)
                self.assertEqual(self.connection.flush(), 0)

    def test_flush_would_block_ssl_write_buffer(self):
        '''ssl.SSL_ERROR_WANT_WRITE usesthe same buffer on next send'''
        pending = deque(map(str, [1, 2, 3]))
        with mock.patch.object(self.connection, '_pending', pending):
            with mock.patch.object(self.connection, '_socket') as mock_socket:
                mock_socket.send.side_effect = ssl.SSLError(
                    ssl.SSL_ERROR_WANT_WRITE)
                self.assertFalse(self.connection._out_buffer)
                self.connection.flush()
                self.assertEqual(self.connection._out_buffer, '123')

        # With some more pending items, make sure we still only get '123' sent
        pending = deque(map(str, [4, 5, 6]))
        with mock.patch.object(self.connection, '_pending', pending):
            with mock.patch.object(self.connection, '_socket') as mock_socket:
                mock_socket.send.return_value = 3
                # The first flush should see the existing buffer
                self.connection.flush()
                mock_socket.send.assert_called_with('123')
                # The second flush should see the pending requests
                self.connection.flush()
                mock_socket.send.assert_called_with('456')

    def test_flush_socket_error(self):
        '''Re-raises socket non-EAGAIN errors'''
        pending = deque(map(str, [1, 2, 3]))
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            with mock.patch.object(self.connection, '_pending', pending):
                mock_socket.send.side_effect = socket.error('foo')
                self.assertRaises(socket.error, self.connection.flush)

    def test_eager_flush(self):
        '''Sending on a non-blocking connection does not eagerly flushes'''
        with mock.patch.object(self.connection, 'flush') as mock_flush:
            self.connection.send('foo')
            mock_flush.assert_not_called()

    def test_close_flush(self):
        '''Closing the connection flushes all remaining messages'''
        def fake_flush():
            self.connection._pending = False
            self.connection._fake_flush_called = True

        with mock.patch.object(self.connection, 'flush', fake_flush):
            self.connection.send('foo')
            self.connection.close()
            self.assertTrue(self.connection._fake_flush_called)

    def test_magic(self):
        '''Sends the NSQ magic bytes'''
        self.assertTrue(self.socket.read().startswith(constants.MAGIC_V2))

    def test_identify(self):
        '''The connection sends the identify commands'''
        expected = ''.join([
            constants.MAGIC_V2,
            constants.IDENTIFY,
            constants.NL,
            util.pack(json.dumps(self.connection._identify_options))])
        self.assertEqual(self.socket.read(), expected)

    def test_read_timeout(self):
        '''Returns no results after a socket timeout'''
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            mock_socket.recv.side_effect = socket.timeout
            self.assertEqual(self.connection.read(), [])

    def test_read_socket_error(self):
        '''Re-raises socket non-errno socket errors'''
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            mock_socket.recv.side_effect = socket.error('foo')
            self.assertRaises(socket.error, self.connection.read)

    def test_read_would_block(self):
        '''Returns no results if it would block'''
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            mock_socket.recv.side_effect = socket.error(errno.EAGAIN)
            self.assertEqual(self.connection.read(), [])

    def test_read_would_block_ssl_write(self):
        '''Returns no results if it would block on a SSL socket'''
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            mock_socket.recv.side_effect = ssl.SSLError(ssl.SSL_ERROR_WANT_WRITE)
            self.assertEqual(self.connection.read(), [])

    def test_read_would_block_ssl_read(self):
        '''Returns no results if it would block on a SSL socket'''
        with mock.patch.object(self.connection, '_socket') as mock_socket:
            mock_socket.recv.side_effect = ssl.SSLError(ssl.SSL_ERROR_WANT_READ)
            self.assertEqual(self.connection.read(), [])

    def test_read_partial(self):
        '''Returns nothing if it has only read partial results'''
        self.socket.write('f')
        self.assertEqual(self.connection.read(), [])

    def test_read_size_partial(self):
        '''Returns one response size is complete, but content is partial'''
        self.socket.write(response.Response.pack('hello')[:-1])
        self.assertEqual(self.connection.read(), [])

    def test_read_whole(self):
        '''Returns a single message if it has read a complete one'''
        self.socket.write(response.Response.pack('hello'))
        expected = response.Response(
            self.connection, constants.FRAME_TYPE_RESPONSE, 'hello')
        self.assertEqual(self.connection.read(), [expected])

    def test_read_multiple(self):
        '''Returns multiple responses if available'''
        self.socket.write(response.Response.pack('hello') * 10)
        expected = response.Response(
            self.connection, constants.FRAME_TYPE_RESPONSE, 'hello')
        self.assertEqual(self.connection.read(), [expected] * 10)

    def test_fileno(self):
        '''Returns the connection's file descriptor appropriately'''
        self.assertEqual(
            self.connection.fileno(), self.socket.fileno())

    def test_fileno_closed(self):
        '''Raises an exception if the connection's closed'''
        with mock.patch.object(self.connection, '_socket', None):
            self.assertRaises(exceptions.ConnectionClosedException,
                self.connection.fileno)

    def test_str_alive(self):
        '''Sane str representation for an alive connection'''
        with mock.patch.object(self.connection, 'alive', return_value=True):
            with mock.patch.object(
                self.connection, 'fileno', return_value=7):
                with mock.patch.object(self.connection, 'host', 'host'):
                    with mock.patch.object(self.connection, 'port', 'port'):
                        self.assertEqual(str(self.connection),
                            '<Connection host:port (alive on FD 7)>')

    def test_str_dead(self):
        '''Sane str representation for an alive connection'''
        with mock.patch.object(self.connection, 'alive', return_value=False):
            with mock.patch.object(
                self.connection, 'fileno', return_value=7):
                with mock.patch.object(self.connection, 'host', 'host'):
                    with mock.patch.object(self.connection, 'port', 'port'):
                        self.assertEqual(str(self.connection),
                            '<Connection host:port (dead on FD 7)>')

    def test_send_no_message(self):
        '''Appropriately sends packed data without message'''
        self.socket.read()
        self.connection.nop()
        self.connection.flush()
        expected = constants.NOP + constants.NL
        self.assertEqual(self.socket.read(), expected)

    def test_send_message(self):
        '''Appropriately sends packed data with message'''
        self.socket.read()
        self.connection.identify({})
        self.connection.flush()
        expected = ''.join(
            (constants.IDENTIFY, constants.NL, util.pack('{}')))
        self.assertEqual(self.socket.read(), expected)

    def assertSent(self, expected, function, *args, **kwargs):
        '''Assert that the connection sends the expected payload'''
        self.socket.read()
        function(*args, **kwargs)
        self.connection.flush()
        self.assertEqual(self.socket.read(), expected)

    def test_auth(self):
        '''Appropriately send auth'''
        expected = ''.join((constants.AUTH, constants.NL, util.pack('hello')))
        self.assertSent(expected, self.connection.auth, 'hello')

    def test_sub(self):
        '''Appropriately sends sub'''
        expected = ''.join((constants.SUB, ' foo bar', constants.NL))
        self.assertSent(expected, self.connection.sub, 'foo', 'bar')

    def test_pub(self):
        '''Appropriately sends pub'''
        expected = ''.join(
            (constants.PUB, ' foo', constants.NL, util.pack('hello')))
        self.assertSent(expected, self.connection.pub, 'foo', 'hello')

    def test_mpub(self):
        '''Appropriately sends mpub'''
        expected = ''.join((
            constants.MPUB, ' foo', constants.NL,
            util.pack(['hello', 'howdy'])))
        self.assertSent(expected, self.connection.mpub, 'foo', 'hello', 'howdy')

    def test_ready(self):
        '''Appropriately sends ready'''
        expected = ''.join((constants.RDY, ' 5', constants.NL))
        self.assertSent(expected, self.connection.rdy, 5)

    def test_fin(self):
        '''Appropriately sends fin'''
        expected = ''.join((constants.FIN, ' message_id', constants.NL))
        self.assertSent(expected, self.connection.fin, 'message_id')

    def test_req(self):
        '''Appropriately sends req'''
        expected = ''.join((constants.REQ, ' message_id 10', constants.NL))
        self.assertSent(expected, self.connection.req, 'message_id', 10)

    def test_touch(self):
        '''Appropriately sends touch'''
        expected = ''.join((constants.TOUCH, ' message_id', constants.NL))
        self.assertSent(expected, self.connection.touch, 'message_id')

    def test_cls(self):
        '''Appropriately sends cls'''
        expected = ''.join((constants.CLS, constants.NL))
        self.assertSent(expected, self.connection.cls)

    def test_nop(self):
        '''Appropriately sends nop'''
        expected = ''.join((constants.NOP, constants.NL))
        self.assertSent(expected, self.connection.nop)

    # Some tests very closely aimed at identification
    def test_calls_identified(self):
        '''Upon getting an identification response, we call 'identified'''
        with mock.patch.object(
            connection.Connection, 'identified') as mock_identified:
            self.connect({'foo': 'bar'})
            self.assertTrue(mock_identified.called)

    def test_identified_tolerates_ok(self):
        '''The identified handler tolerates OK responses'''
        res = mock.Mock(data='OK')
        self.assertEqual(self.connection.identified(res).data, 'OK')

    def test_identify_defaults(self):
        '''Identify provides default options'''
        self.assertEqual(self.connection._identify_options, {
            'feature_negotiation': True,
            'long_id': socket.getfqdn(),
            'short_id': socket.gethostname(),
            'user_agent': self.connection.USER_AGENT
        })

    def test_identify_override_defaults(self):
        '''Identify allows us to override defaults'''
        with mock.patch('nsq.connection.Connection.connect'):
            conn = connection.Connection('host', 0, long_id='not-your-fqdn')
            self.assertEqual(conn._identify_options['long_id'], 'not-your-fqdn')

    def test_identify_tls_unsupported(self):
        '''Raises an exception about the lack of TLS support'''
        with mock.patch('nsq.connection.TLSSocket', None):
            self.assertRaises(exceptions.UnsupportedException,
                connection.Connection, 'host', 0, tls_v1=True)

    def test_identify_snappy_unsupported(self):
        '''Raises an exception about the lack of snappy support'''
        with mock.patch('nsq.connection.SnappySocket', None):
            self.assertRaises(exceptions.UnsupportedException,
                connection.Connection, 'host', 0, snappy=True)

    def test_identify_deflate_unsupported(self):
        '''Raises an exception about the lack of deflate support'''
        with mock.patch('nsq.connection.DeflateSocket', None):
            self.assertRaises(exceptions.UnsupportedException,
                connection.Connection, 'host', 0, deflate=True)

    def test_identify_no_deflate_level(self):
        '''Raises an exception about the lack of deflate_level support'''
        with mock.patch('nsq.connection.DeflateSocket', None):
            self.assertRaises(exceptions.UnsupportedException,
                connection.Connection, 'host', 0, deflate_level=True)

    def test_identify_no_snappy_and_deflate(self):
        '''We should yell early about incompatible snappy and deflate options'''
        self.assertRaises(exceptions.UnsupportedException,
            connection.Connection, 'host', 0, snappy=True, deflate=True)

    def test_identify_saves_identify_response(self):
        '''Saves the identify response from the server'''
        expected = {'foo': 'bar'}
        conn = self.connect(expected)
        self.assertEqual(conn._identify_response, expected)

    def test_identify_saves_max_rdy_count(self):
        '''Saves the max ready count if it's provided'''
        conn = self.connect({'max_rdy_count': 100})
        self.assertEqual(conn.max_rdy_count, 100)

    def test_ready_to_reconnect(self):
        '''Alias for the reconnection attempt's ready method'''
        with mock.patch.object(
            self.connection, '_reconnnection_counter') as ctr:
            self.connection.ready_to_reconnect()
            ctr.ready.assert_called_with()

    def test_reconnect_living_socket(self):
        '''Don't reconnect a living connection'''
        before = self.connection._socket
        self.connection.connect()
        self.assertEqual(self.connection._socket, before)

    def test_connect_socket_error_return_value(self):
        '''Socket errors has connect return False'''
        self.connection.close()
        with mock.patch('nsq.connection.socket') as mock_socket:
            mock_socket.socket = mock.Mock(side_effect=socket.error)
            self.assertFalse(self.connection.connect())

    def test_connect_socket_error_reset(self):
        '''Invokes reset if the socket raises an error'''
        self.connection.close()
        with mock.patch('nsq.connection.socket') as mock_socket:
            with mock.patch.object(self.connection, '_reset') as mock_reset:
                mock_socket.socket = mock.Mock(side_effect=socket.error)
                self.connection.connect()
                mock_reset.assert_called_with()

    def test_connect_timeout(self):
        '''Times out when connection instantiation is too slow'''
        socket = self.connection._socket
        self.connection.close()
        with mock.patch.object(self.connection, '_read', return_value=[]):
            with mock.patch.object(self.connection, '_timeout', 0.05):
                with mock.patch(
                    'nsq.connection.socket.socket', return_value=socket):
                    self.assertFalse(self.connection.connect())

    def test_connect_resets_state(self):
        '''Upon connection, makes a call to reset its state'''
        socket = self.connection._socket
        self.connection.close()
        with mock.patch.object(self.connection, '_read', return_value=[]):
            with mock.patch.object(self.connection, '_reset') as mock_reset:
                with mock.patch.object(self.connection, '_timeout', 0.05):
                    with mock.patch(
                        'nsq.connection.socket.socket', return_value=socket):
                        self.connection.connect()
                        mock_reset.assert_called_with()

    def test_close_resets_state(self):
        '''On closing a connection, reset its state'''
        with mock.patch.object(self.connection, '_reset') as mock_reset:
            self.connection.close()
            mock_reset.assert_called_with()

    def test_reset_socket(self):
        '''Resets socket'''
        self.connection._socket = True
        self.connection._reset()
        self.assertEqual(self.connection._socket, None)

    def test_reset_pending(self):
        '''Resets pending'''
        self.connection._pending = True
        self.connection._reset()
        self.assertEqual(self.connection._pending, deque())

    def test_reset_out_buffer(self):
        '''Resets the outbound buffer'''
        self.connection._out_buffer = True
        self.connection._reset()
        self.assertEqual(self.connection._out_buffer, '')

    def test_reset_buffer(self):
        '''Resets buffer'''
        self.connection._buffer = True
        self.connection._reset()
        self.assertEqual(self.connection._buffer, '')

    def test_reset_identify_response(self):
        '''Resets identify_response'''
        self.connection._identify_response = True
        self.connection._reset()
        self.assertEqual(self.connection._identify_response, {})

    def test_reset_last_ready_sent(self):
        '''Resets last_ready_sent'''
        self.connection.last_ready_sent = True
        self.connection._reset()
        self.assertEqual(self.connection.last_ready_sent, 0)

    def test_reset_ready(self):
        '''Resets ready'''
        self.connection.ready = True
        self.connection._reset()
        self.assertEqual(self.connection.ready, 0)

    def test_ok_response(self):
        '''Sets our _identify_response to {} if 'OK' is provided'''
        res = response.Response(
            self.connection, response.Response.FRAME_TYPE, 'OK')
        self.connection.identified(res)
        self.assertEqual(self.connection._identify_response, {})

    def test_tls_unsupported(self):
        '''Raises an exception if the server does not support TLS'''
        res = response.Response(self.connection,
            response.Response.FRAME_TYPE, json.dumps({'tls_v1': False}))
        options = {'tls_v1': True}
        with mock.patch.object(self.connection, '_identify_options', options):
            self.assertRaises(exceptions.UnsupportedException,
                self.connection.identified, res)

    def test_auth_required_not_provided(self):
        '''Raises an exception if auth is required but not provided'''
        res = response.Response(self.connection, response.Response.FRAME_TYPE,
            json.dumps({'auth_required': True}))
        self.assertRaises(exceptions.UnsupportedException,
            self.connection.identified, res)

    def test_auth_required_provided(self):
        '''Sends the auth message if required and provided'''
        res = response.Response(self.connection, response.Response.FRAME_TYPE,
            json.dumps({'auth_required': True}))
        with mock.patch.object(self.connection, 'auth') as mock_auth:
            with mock.patch.object(self.connection, '_auth_secret', 'hello'):
                self.connection.identified(res)
                mock_auth.assert_called_with('hello')

    def test_auth_provided_not_required(self):
        '''Logs a warning if you provide auth when none is required'''
        res = response.Response(self.connection, response.Response.FRAME_TYPE,
            json.dumps({'auth_required': False}))
        with mock.patch('nsq.connection.logger') as mock_logger:
            with mock.patch.object(self.connection, '_auth_secret', 'hello'):
                self.connection.identified(res)
                mock_logger.warn.assert_called_with(
                    'Authentication secret provided but not required')


class TestTLSConnectionIntegration(HttpClientIntegrationTest):
    '''We can establish a connection with TLS'''
    def setUp(self):
        HttpClientIntegrationTest.setUp(self)
        self.connection = connection.Connection('localhost', 14150, tls_v1=True)
        self.connection.setblocking(0)

    def test_alive(self):
        '''The connection is alive'''
        self.assertTrue(self.connection.alive())

    def test_basic(self):
        '''Can send and receive things'''
        self.connection.pub('foo', 'bar')
        self.connection.flush()
        responses = []
        while not responses:
            responses = self.connection.read()
        self.assertEqual(len(responses), 1)
        self.assertEqual(responses[0].data, 'OK')



'''Some utilities used around town'''

import struct


def pack_string(message):
    '''Pack a single message in the TCP protocol format'''
    # [ 4-byte message size ][ N-byte binary data ]
    return struct.pack('>l', len(message)) + message


def pack_iterable(messages):
    '''Pack an iterable of messages in the TCP protocol format'''
    # [ 4-byte body size ]
    # [ 4-byte num messages ]
    # [ 4-byte message #1 size ][ N-byte binary data ]
    #      ... (repeated <num_messages> times)
    return pack_string(
        struct.pack('>l', len(messages)) +
        ''.join(map(pack_string, messages)))


def pack(message):
    '''Pack the provided message'''
    if isinstance(message, basestring):
        return pack_string(message)
    else:
        return pack_iterable(message)


def hexify(message):
    '''Print out printable characters, but others in hex'''
    import string
    hexified = []
    for char in message:
        if (char in '\n\r \t') or (char not in string.printable):
            hexified.append('\\x%02x' % ord(char))
        else:
            hexified.append(char)
    return ''.join(hexified)


def distribute(total, objects):
    '''Generator for (count, object) tuples that distributes count evenly among
    the provided objects'''
    for index, obj in enumerate(objects):
        start = (index * total) / len(objects)
        stop = ((index + 1) * total) / len(objects)
        yield (stop - start, obj)



'''Classes that know about backoffs'''

import sys
import time


class Backoff(object):
    '''Backoff base class'''
    def sleep(self, attempt):
        '''Sleep for the duration of this backoff'''
        time.sleep(self.backoff(attempt))

    def backoff(self, attempt):
        '''The amount of time this attempt requires'''
        raise NotImplementedError()


class Linear(Backoff):
    '''Linear backoff'''
    def __init__(self, a, b):
        Backoff.__init__(self)
        self._a = a
        self._b = b

    def backoff(self, attempt):
        return self._a * attempt + self._b


class Constant(Linear):
    '''Always the same backoff'''
    def __init__(self, constant):
        Linear.__init__(self, 0, constant)


class Exponential(Backoff):
    '''Exponential backoff of the form a * base ^ attempt + c'''
    def __init__(self, base, a=1, c=0):
        Backoff.__init__(self)
        self._base = base
        self._a = a
        self._c = c

    def backoff(self, attempt):
        return self._a * (self._base ** attempt) + self._c


class Clamped(Backoff):
    '''Backoff clamped to min / max bounds'''
    def __init__(self, backoff, minimum=0, maximum=sys.maxint):
        Backoff.__init__(self)
        self._min = minimum
        self._max = maximum
        self._backoff = backoff

    def backoff(self, attempt):
        return max(self._min, min(self._max, self._backoff.backoff(attempt)))


class AttemptCounter(object):
    '''Count the number of attempts we've used'''
    def __init__(self, backoff):
        self.attempts = 0
        self._backoff = backoff
        self._last_failed = None

    def sleep(self):
        '''Sleep for the duration of this backoff'''
        time.sleep(self.backoff())

    def backoff(self):
        '''Get the current backoff'''
        return self._backoff.backoff(self.attempts)

    def success(self):
        '''Update the attempts count correspondingly'''
        self._last_failed = None

    def failed(self):
        '''Update the attempts count correspondingly'''
        self._last_failed = time.time()
        self.attempts += 1

    def ready(self):
        '''Whether or not enough time has passed since the last failure'''
        if self._last_failed:
            delta = time.time() - self._last_failed
            return delta >= self.backoff()
        return True


class ResettingAttemptCounter(AttemptCounter):
    '''A counter that resets on success'''
    def success(self):
        AttemptCounter.success(self)
        self.attempts = 0


class DecrementingAttemptCounter(AttemptCounter):
    '''A counter that decrements attempts on success'''
    def success(self):
        AttemptCounter.success(self)
        self.attempts = max(0, self.attempts - 1)



import unittest

import mock
import uuid
import socket
import struct
from nsq import response
from nsq import constants
from nsq import exceptions


class TestResponse(unittest.TestCase):
    '''Test our response class'''
    def test_from_raw_response(self):
        '''Make sure we can construct a raw response'''
        raw = struct.pack('>l5s', constants.FRAME_TYPE_RESPONSE, 'hello')
        res = response.Response.from_raw(None, raw)
        self.assertEqual(res.__class__, response.Response)
        self.assertEqual(res.data, 'hello')

    def test_from_raw_unknown_frame(self):
        '''Raises an exception for unknown frame types'''
        raw = struct.pack('>l5s', 9042, 'hello')
        self.assertRaises(TypeError, response.Response.from_raw, None, raw)

    def test_str(self):
        '''Has a reasonable string value'''
        raw = struct.pack('>l5s', constants.FRAME_TYPE_RESPONSE, 'hello')
        res = response.Response.from_raw(None, raw)
        self.assertEqual(str(res), 'Response - hello')

    def test_pack(self):
        '''Can pack itself up'''
        packed = response.Response.pack('hello')[4:]
        unpacked = response.Response.from_raw(None, packed)
        self.assertIsInstance(unpacked, response.Response)
        self.assertEqual(unpacked.data, 'hello')


class TestError(unittest.TestCase):
    '''Test our error response class'''
    def test_from_raw_error(self):
        '''Can identify an error type'''
        raw = struct.pack('>l5s', constants.FRAME_TYPE_ERROR, 'hello')
        res = response.Response.from_raw(None, raw)
        self.assertEqual(res.__class__, response.Error)
        self.assertEqual(res.data, 'hello')

    def test_str(self):
        '''Has a reasonable string value'''
        raw = struct.pack('>l5s', constants.FRAME_TYPE_ERROR, 'hello')
        res = response.Response.from_raw(None, raw)
        self.assertEqual(str(res), 'Error - hello')

    def test_find(self):
        '''Can correctly identify the appropriate exception'''
        expected = {
            'E_INVALID': exceptions.InvalidException,
            'E_BAD_BODY': exceptions.BadBodyException,
            'E_BAD_TOPIC': exceptions.BadTopicException,
            'E_BAD_CHANNEL': exceptions.BadChannelException,
            'E_BAD_MESSAGE': exceptions.BadMessageException,
            'E_PUB_FAILED': exceptions.PubFailedException,
            'E_MPUB_FAILED': exceptions.MpubFailedException,
            'E_FIN_FAILED': exceptions.FinFailedException,
            'E_REQ_FAILED': exceptions.ReqFailedException,
            'E_TOUCH_FAILED': exceptions.TouchFailedException
        }
        for key, klass in expected.items():
            self.assertEqual(response.Error.find(key), klass)

    def test_find_missing(self):
        '''Raises an exception if it can't find the appropriate exception'''
        self.assertRaises(TypeError, response.Error.find, 'woruowijklf')

    def test_exception(self):
        '''Can correctly raise the appropriate exception'''
        raw = struct.pack('>l13s', constants.FRAME_TYPE_ERROR, 'E_INVALID foo')
        res = response.Response.from_raw(None, raw)
        exc = res.exception()
        self.assertIsInstance(exc, exceptions.InvalidException)
        self.assertEqual(exc.message, 'foo')

    def test_pack(self):
        '''Can pack itself up'''
        packed = response.Error.pack('hello')[4:]
        unpacked = response.Response.from_raw(None, packed)
        self.assertIsInstance(unpacked, response.Error)
        self.assertEqual(unpacked.data, 'hello')


class TestMessage(unittest.TestCase):
    '''Test our message case'''
    def setUp(self):
        self.id = uuid.uuid4().hex[:16]
        self.timestamp = 0
        self.attempt = 1
        self.body = 'hello'
        self.packed = struct.pack('>qH16s5s', 0, 1, self.id, self.body)
        self.response = response.Response.from_raw(mock.Mock(),
            struct.pack('>l31s', constants.FRAME_TYPE_MESSAGE, self.packed))

    def test_str(self):
        '''Has a reasonable string value'''
        self.assertEqual(str(self.response), 'Message - 0 1 %s hello' % self.id)

    def test_from_raw_message(self):
        '''Can identify a message type'''
        self.assertEqual(self.response.__class__, response.Message)

    def test_timestamp(self):
        '''Can identify the timestamp'''
        self.assertEqual(self.response.timestamp, self.timestamp)

    def test_attempts(self):
        '''Can identify the number of attempts'''
        self.assertEqual(self.response.attempts, self.attempt)

    def test_id(self):
        '''Can identify the ID of the message'''
        self.assertEqual(self.response.id, self.id)

    def test_message(self):
        '''Can properly detect the message'''
        self.assertEqual(self.response.body, self.body)

    def test_fin(self):
        '''Invokes the fin method'''
        self.response.fin()
        self.response.connection.fin.assert_called_with(self.id)

    def test_req(self):
        '''Invokes the req method'''
        self.response.req(1)
        self.response.connection.req.assert_called_with(self.id, 1)

    def test_touch(self):
        '''Invokes the touch method'''
        self.response.touch()
        self.response.connection.touch.assert_called_with(self.id)

    def test_pack(self):
        '''Can pack itself up'''
        packed = response.Message.pack(
            self.timestamp, self.attempt, self.id, self.body)[4:]
        unpacked = response.Response.from_raw(None, packed)
        self.assertIsInstance(unpacked, response.Message)
        self.assertEqual(unpacked.timestamp, self.timestamp)
        self.assertEqual(unpacked.attempts, self.attempt)
        self.assertEqual(unpacked.id, self.id)
        self.assertEqual(unpacked.body, self.body)

    def test_handle_yields(self):
        '''The handle method should yield the message'''
        with self.response.handle() as msg:
            self.assertEqual(msg, self.response)

    def test_handle_exception(self):
        '''Handles exceptions by requeueing'''
        try:
            with self.response.handle():
                raise ValueError('foo')
        except ValueError:
            self.response.connection.req.assert_called_with(
                self.response.id, self.response.delay())
        else:
            self.assertTrue(False, 'No exception was raised')

    def test_handle_success(self):
        '''Handles success by calling fin'''
        with self.response.handle():
            pass
        self.response.connection.fin.assert_called_with(self.response.id)

    def test_handle_already_requeued(self):
        '''If we've already requeued a message, doesn't requeue it again'''
        try:
            with self.response.handle():
                self.response.req(10)
                raise ValueError('foo')
        except ValueError:
            self.assertEqual(self.response.connection.req.call_count, 1)

    def test_handle_already_finish(self):
        '''If we've already finished a messages, doesn't finish it again'''
        with self.response.handle():
            self.response.fin()
        self.assertEqual(self.response.connection.fin.call_count, 1)

    def test_handle_exception_socket_error(self):
        '''Handles socket errors when catching exceptions'''
        try:
            self.response.connection.req = mock.Mock(side_effect=socket.error)
            with self.response.handle():
                raise ValueError('foo')
        except ValueError:
            # The connection should have been closed
            self.response.connection.close.assert_called_with()

    def test_handle_success_socket_error(self):
        '''Handles socket errors when trying to complete the message'''
        try:
            self.response.connection.fin = mock.Mock(side_effect=socket.error)
            with self.response.handle():
                pass
        except ValueError:
            # The connection should have been closed
            self.response.connection.close.assert_called_with()



'''Tests about our connection-checking thread'''

import mock
import unittest
import threading

from nsq.checker import PeriodicThread, ConnectionChecker


class TestPeriodicThread(unittest.TestCase):
    '''Can stop a PeriodicThread'''
    def test_stop(self):
        '''Stop is effective for stopping a stoppable thread'''
        def callback(event):
            '''Trigger an event'''
            event.set()

        event = threading.Event()
        thread = PeriodicThread(0.01, callback, event)
        thread.start()
        event.wait()
        thread.stop()
        thread.join()
        self.assertFalse(thread.is_alive())

    def test_repeats(self):
        '''Repeats the same callback several times'''
        def callback(event, counter):
            '''Trigger an event after accumulating enough'''
            counter['count'] += 1
            if counter['count'] > 10:
                event.set()

        event = threading.Event()
        counter = {'count': 0}
        thread = PeriodicThread(0.01, callback, event, counter)
        thread.start()
        event.wait()
        thread.stop()
        thread.join()
        self.assertGreaterEqual(counter['count'], 10)

    def test_survives_standard_error(self):
        '''The thread survives exceptions'''
        def callback():
            '''Raise an exception'''
            raise StandardError('foo')

        with mock.patch('nsq.checker.logger') as mock_logger:
            thread = PeriodicThread(0.01, callback)
            thread.start()
            thread.join(0.1)
            self.assertTrue(thread.is_alive())
            thread.stop()
            thread.join()
            mock_logger.exception.assert_called_with('Callback failed')
            self.assertGreater(mock_logger.exception.call_count, 1)


class TestConnectionChecker(unittest.TestCase):
    '''ConnectionChecker tests'''
    def test_callback(self):
        '''Provides the client's connection checking method'''
        with mock.patch.object(PeriodicThread, '__init__') as mock_init:
            mock_client = mock.Mock()
            checker = ConnectionChecker(mock_client, 10)
            mock_init.assert_called_with(
                checker, 10, mock_client.check_connections)



'''A class for interacting with a nsqd instance over http'''

from . import BaseClient, json_wrap, ok_check, ClientException
from ..util import pack


class Client(BaseClient):
    @ok_check
    def ping(self):
        '''Ping the client'''
        return self.get('ping')

    @json_wrap
    def info(self):
        '''Get information about the client'''
        return self.get('info')

    @ok_check
    def pub(self, topic, message):
        '''Publish a message to a topic'''
        return self.post('pub', params={'topic': topic}, data=message)

    @ok_check
    def mpub(self, topic, messages, binary=True):
        '''Send multiple messages to a topic. Optionally pack the messages'''
        if binary:
            # Pack and ship the data
            return self.post('mpub', data=pack(messages)[4:],
                params={'topic': topic, 'binary': True})
        elif any('\n' in m for m in messages):
            # If any of the messages has a newline, then you must use the binary
            # calling format
            raise ClientException(
                'Use `binary` flag in mpub for messages with newlines')
        else:
            return self.post(
                '/mpub', params={'topic': topic}, data='\n'.join(messages))

    @json_wrap
    def create_topic(self, topic):
        '''Create the provided topic'''
        return self.get('create_topic', params={'topic': topic})

    @json_wrap
    def empty_topic(self, topic):
        '''Empty the provided topic'''
        return self.get('empty_topic', params={'topic': topic})

    @json_wrap
    def delete_topic(self, topic):
        '''Delete the provided topic'''
        return self.get('delete_topic', params={'topic': topic})

    @json_wrap
    def pause_topic(self, topic):
        '''Pause the provided topic'''
        return self.get('pause_topic', params={'topic': topic})

    @json_wrap
    def unpause_topic(self, topic):
        '''Unpause the provided topic'''
        return self.get('unpause_topic', params={'topic': topic})

    @json_wrap
    def create_channel(self, topic, channel):
        '''Create the channel in the provided topic'''
        return self.get(
            '/create_channel', params={'topic': topic, 'channel': channel})

    @json_wrap
    def empty_channel(self, topic, channel):
        '''Empty the channel in the provided topic'''
        return self.get(
            '/empty_channel', params={'topic': topic, 'channel': channel})

    @json_wrap
    def delete_channel(self, topic, channel):
        '''Delete the channel in the provided topic'''
        return self.get(
            '/delete_channel', params={'topic': topic, 'channel': channel})

    @json_wrap
    def pause_channel(self, topic, channel):
        '''Pause the channel in the provided topic'''
        return self.get(
            '/pause_channel', params={'topic': topic, 'channel': channel})

    @json_wrap
    def unpause_channel(self, topic, channel):
        '''Unpause the channel in the provided topic'''
        return self.get(
            '/unpause_channel', params={'topic': topic, 'channel': channel})

    @json_wrap
    def stats(self):
        '''Get stats about the server'''
        return self.get('stats', params={'format': 'json'})

    def clean_stats(self):
        '''Stats with topics and channels keyed on topic and channel names'''
        stats = self.stats()
        if 'topics' in stats:  # pragma: no branch
            topics = stats['topics']
            topics = dict((t.pop('topic_name'), t) for t in topics)
            for topic, data in topics.items():
                if 'channels' in data:  # pragma: no branch
                    channels = data['channels']
                    channels = dict(
                        (c.pop('channel_name'), c) for c in channels)
                    data['channels'] = channels
            stats['topics'] = topics
        return stats



from . import backoff
from . import constants
from . import logger
from . import util
from . import json
from . import __version__
from .exceptions import (
    UnsupportedException, ConnectionClosedException, ConnectionTimeoutException)
from .sockets import TLSSocket, SnappySocket, DeflateSocket
from .response import Response, Message

import errno
import socket
import ssl
import struct
import sys
import time
import threading
from collections import deque


class Connection(object):
    '''A socket-based connection to a NSQ server'''
    # Default user agent
    USER_AGENT = 'nsq-py/%s' % __version__
    # Errors that would block
    WOULD_BLOCK_ERRS = (
        errno.EAGAIN, ssl.SSL_ERROR_WANT_WRITE, ssl.SSL_ERROR_WANT_READ)

    def __init__(self, host, port, timeout=None, reconnection_backoff=None,
        auth_secret=None, **identify):
        assert isinstance(host, (str, unicode))
        assert isinstance(port, int)

        self._reset()

        # Our host and port
        self.host = host
        self.port = port
        # Whether or not our socket is set to block
        self._blocking = 1
        self._timeout = timeout if timeout is not None else 1.0
        # The options to use when identifying
        self._identify_options = dict(identify)
        self._identify_options.setdefault('short_id', socket.gethostname())
        self._identify_options.setdefault('long_id', socket.getfqdn())
        self._identify_options.setdefault('feature_negotiation', True)
        self._identify_options.setdefault('user_agent', self.USER_AGENT)

        # In support of auth
        self._auth_secret = auth_secret

        # Some settings that may be determined by an identify response
        self.max_rdy_count = sys.maxint

        # Check for any options we don't support
        disallowed = []
        if not SnappySocket:  # pragma: no branch
            disallowed.append('snappy')
        if not DeflateSocket:  # pragma: no branch
            disallowed.extend(['deflate', 'deflate_level'])
        if not TLSSocket:  # pragma: no branch
            disallowed.append('tls_v1')
        for key in disallowed:
            if self._identify_options.get(key, False):
                raise UnsupportedException('Option %s is not supported' % key)

        # Our backoff policy for reconnection. The default is to use an
        # exponential backoff 8 * (2 ** attempt) clamped to [0, 60]
        self._reconnection_backoff = (
            reconnection_backoff or
            backoff.Clamped(backoff.Exponential(2, 8), maximum=60))
        self._reconnnection_counter = backoff.ResettingAttemptCounter(
            self._reconnection_backoff)

        # A lock around our socket
        self._socket_lock = threading.RLock()

        # Establish our connection
        self.connect()

    def __str__(self):
        state = 'alive' if self.alive() else 'dead'
        return '<Connection %s:%s (%s on FD %s)>' % (
            self.host, self.port, state, self.fileno())

    def ready_to_reconnect(self):
        '''Returns True if enough time has passed to attempt a reconnection'''
        return self._reconnnection_counter.ready()

    def _reset(self):
        '''Reset all of our stateful variables'''
        self._socket = None
        # The pending messages we have to send, and the current buffer we're
        # sending
        self._pending = deque()
        self._out_buffer = ''
        # Our read buffer
        self._buffer = ''
        # The identify response we last received from the server
        self._identify_response = {}
        # Our ready state
        self.last_ready_sent = 0
        self.ready = 0

    def connect(self, force=False):
        '''Establish a connection'''
        # Don't re-establish existing connections
        if not force and self.alive():
            return True

        self._reset()

        # Otherwise, try to connect
        with self._socket_lock:
            try:
                logger.info('Creating socket...')
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._socket.settimeout(self._timeout)
                logger.info('Connecting to %s, %s', self.host, self.port)
                self._socket.connect((self.host, self.port))
                # Set our socket's blocking state to whatever ours is
                self._socket.setblocking(self._blocking)
                # Safely write our magic
                self._pending.append(constants.MAGIC_V2)
                while self.pending():
                    self.flush()
                # And send our identify command
                self.identify(self._identify_options)
                while self.pending():
                    self.flush()
                self._reconnnection_counter.success()
                # Wait until we've gotten a response to IDENTIFY, try to read
                # one. Also, only spend up to the provided timeout waiting to
                # establish the connection.
                limit = time.time() + self._timeout
                responses = self._read(1)
                while (not responses) and (time.time() < limit):
                    responses = self._read(1)
                if not responses:
                    raise ConnectionTimeoutException(
                        'Read identify response timed out (%ss)' % self._timeout)
                self.identified(responses[0])
                return True
            except:
                logger.exception('Failed to connect')
                if self._socket:
                    self._socket.close()
                self._reconnnection_counter.failed()
                self._reset()
                return False

    def close(self):
        '''Close our connection'''
        # Flush any unsent message
        try:
            while self.pending():
                self.flush()
        except socket.error:
            pass
        with self._socket_lock:
            try:
                if self._socket:
                    self._socket.close()
            finally:
                self._reset()

    def socket(self, blocking=True):
        '''Blockingly yield the socket'''
        # If the socket is available, then yield it. Otherwise, yield nothing
        if self._socket_lock.acquire(blocking):
            try:
                yield self._socket
            finally:
                self._socket_lock.release()

    def identified(self, res):
        '''Handle a response to our 'identify' command. Returns response'''
        # If they support it, they should give us a JSON blob which we should
        # inspect.
        try:
            res.data = json.loads(res.data)
            self._identify_response = res.data
            logger.info('Got identify response: %s', res.data)
        except:
            logger.warn('Server does not support feature negotiation')
            self._identify_response = {}

        # Save our max ready count unless it's not provided
        self.max_rdy_count = self._identify_response.get(
            'max_rdy_count', self.max_rdy_count)
        if self._identify_options.get('tls_v1', False):
            if not self._identify_response.get('tls_v1', False):
                raise UnsupportedException(
                    'NSQd instance does not support TLS')
            else:
                self._socket = TLSSocket.wrap_socket(self._socket)

        # Now is the appropriate time to send auth
        if self._identify_response.get('auth_required', False):
            if not self._auth_secret:
                raise UnsupportedException(
                    'Auth required but not provided')
            else:
                self.auth(self._auth_secret)
                # If we're not talking over TLS, warn the user
                if not self._identify_response.get('tls_v1', False):
                    logger.warn('Using AUTH without TLS')
        elif self._auth_secret:
            logger.warn('Authentication secret provided but not required')
        return res

    def alive(self):
        '''Returns True if this connection is alive'''
        return bool(self._socket)

    def setblocking(self, blocking):
        '''Set whether or not this message is blocking'''
        for sock in self.socket():
            sock.setblocking(blocking)
            self._blocking = blocking

    def fileno(self):
        '''Returns the socket's fileno. This allows us to select on this'''
        for sock in self.socket():
            if sock:
                return sock.fileno()
        raise ConnectionClosedException()

    def pending(self):
        '''All of the messages waiting to be sent'''
        return self._pending

    def flush(self):
        '''Flush some of the waiting messages, returns count written'''
        # When profiling, we found that while there was some efficiency to be
        # gained elsewhere, the big performance hit is sending lots of small
        # messages at a time. In particular, consumers send many 'FIN' messages
        # which are very small indeed and the cost of dispatching so many system
        # calls is very high. Instead, we prefer to glom together many messages
        # into a single string to send at once.
        total = 0
        for sock in self.socket(blocking=False):
            # If there's nothing left in the out buffer, take whatever's in the
            # pending queue.
            #
            # When using SSL, if the socket throws 'SSL_WANT_WRITE', then the
            # subsequent send requests have to send the same buffer.
            pending = self._pending
            data = self._out_buffer or ''.join(
                pending.popleft() for _ in xrange(len(pending)))
            try:
                # Try to send as much of the first message as possible
                total = sock.send(data)
            except socket.error as exc:
                # Catch (errno, message)-type socket.errors
                if exc.args[0] not in self.WOULD_BLOCK_ERRS:
                    raise
                self._out_buffer = data
            else:
                self._out_buffer = None
            finally:
                if total < len(data):
                    # Save the rest of the message that could not be sent
                    self._pending.appendleft(data[total:])
        return total

    def send(self, command, message=None):
        '''Send a command over the socket with length endcoded'''
        if message:
            joined = command + constants.NL + util.pack(message)
        else:
            joined = command + constants.NL
        if self._blocking:
            for sock in self.socket():
                sock.sendall(joined)
        else:
            self._pending.append(joined)

    def identify(self, data):
        '''Send an identification message'''
        return self.send(constants.IDENTIFY, json.dumps(data))

    def auth(self, secret):
        '''Send an auth secret'''
        return self.send(constants.AUTH, secret)

    def sub(self, topic, channel):
        '''Subscribe to a topic/channel'''
        return self.send(' '.join((constants.SUB, topic, channel)))

    def pub(self, topic, message):
        '''Publish to a topic'''
        return self.send(' '.join((constants.PUB, topic)), message)

    def mpub(self, topic, *messages):
        '''Publish multiple messages to a topic'''
        return self.send(constants.MPUB + ' ' + topic, messages)

    def rdy(self, count):
        '''Indicate that you're ready to receive'''
        self.ready = count
        self.last_ready_sent = count
        return self.send(constants.RDY + ' ' + str(count))

    def fin(self, message_id):
        '''Indicate that you've finished a message ID'''
        return self.send(constants.FIN + ' ' + message_id)

    def req(self, message_id, timeout):
        '''Re-queue a message'''
        return self.send(constants.REQ + ' ' + message_id + ' ' + str(timeout))

    def touch(self, message_id):
        '''Reset the timeout for an in-flight message'''
        return self.send(constants.TOUCH + ' ' + message_id)

    def cls(self):
        '''Close the connection cleanly'''
        return self.send(constants.CLS)

    def nop(self):
        '''Send a no-op'''
        return self.send(constants.NOP)

    # These are the various incarnations of our read method. In some instances,
    # we want to return responses in the typical way. But while establishing
    # connections or negotiating a TLS connection, we need to do different
    # things
    def _read(self, limit=1000):
        '''Return all the responses read'''
        # It's important to know that it may return no responses or multiple
        # responses. It depends on how the buffering works out. First, read from
        # the socket
        for sock in self.socket():
            if sock is None:
                # Race condition. Connection has been closed.
                return []
            try:
                packet = sock.recv(4096)
            except socket.timeout:
                # If the socket times out, return nothing
                return []
            except socket.error as exc:
                # Catch (errno, message)-type socket.errors
                if exc.args[0] in self.WOULD_BLOCK_ERRS:
                    return []
                else:
                    raise

            # Append our newly-read data to our buffer
            self._buffer += packet

        responses = []
        total = 0
        buf = self._buffer
        remaining = len(buf)
        while limit and (remaining >= 4):
            size = struct.unpack('>l', buf[total:(total + 4)])[0]
            # Now check to see if there's enough left in the buffer to read
            # the message.
            if (remaining - 4) >= size:
                responses.append(Response.from_raw(
                    self, buf[(total + 4):(total + size + 4)]))
                total += (size + 4)
                remaining -= (size + 4)
                limit -= 1
            else:
                break
        self._buffer = self._buffer[total:]
        return responses

    def read(self):
        '''Responses from an established socket'''
        responses = self._read()
        # Determine the number of messages in here and decrement our ready
        # count appropriately
        self.ready -= sum(
            map(int, (r.frame_type == Message.FRAME_TYPE for r in responses)))
        return responses



import logging

# Logging, obviously
logger = logging.getLogger('nsq')
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s [%(levelname)s] %(filename)s@%(lineno)d: %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.INFO)

# Our underlying json implmentation
try:
    import simplejson as json
except ImportError:  # pragma: no cover
    import json

# The current version
__version__ = '0.1.11'



'''A client for talking to NSQ'''

from . import connection
from . import logger
from . import exceptions
from .constants import HEARTBEAT
from .response import Response, Error
from .http import nsqlookupd, ClientException
from .checker import ConnectionChecker

from contextlib import contextmanager
import random
import select
import socket
import time
import threading
import math


class Client(object):
    '''A client for talking to NSQ over a connection'''
    def __init__(self,
        lookupd_http_addresses=None, nsqd_tcp_addresses=None, topic=None,
        timeout=0.1, reconnection_backoff=None, auth_secret=None, connect_timeout=None, **identify):
        # If lookupd_http_addresses are provided, so must a topic be.
        if lookupd_http_addresses:
            assert topic

        # Create clients for each of lookupd instances
        lookupd_http_addresses = lookupd_http_addresses or []
        params = {}
        if auth_secret:
            params['access_token'] = auth_secret
        self._lookupd = [
            nsqlookupd.Client(host, **params) for host in lookupd_http_addresses]
        self._topic = topic

        # The select timeout
        self._timeout = timeout
        # Our reconnection backoff policy
        self._reconnection_backoff = reconnection_backoff
        # The connection timeout to pass to the `Connection` class
        self._connect_timeout = connect_timeout

        # The options to send along with identify when establishing connections
        self._identify_options = identify
        self._auth_secret = auth_secret
        # A mapping of (host, port) to our nsqd connection objects
        self._connections = {}

        self._nsqd_tcp_addresses = nsqd_tcp_addresses or []
        self.heartbeat_interval = 30 * 1000
        self.last_recv_timestamp = time.time()
        # A lock for manipulating our connections
        self._lock = threading.RLock()
        # And lastly, instantiate our connections
        self.check_connections()

    def discover(self, topic):
        '''Run the discovery mechanism'''
        logger.info('Discovering on topic %s', topic)
        producers = []
        for lookupd in self._lookupd:
            logger.info('Discovering on %s', lookupd)
            try:
                # Find all the current producers on this instance
                for producer in lookupd.lookup(topic)['producers']:
                    logger.info('Found producer %s on %s', producer, lookupd)
                    producers.append(
                        (producer['broadcast_address'], producer['tcp_port']))
            except ClientException:
                logger.exception('Failed to query %s', lookupd)

        new = []
        for host, port in producers:
            conn = self._connections.get((host, port))
            if not conn:
                logger.info('Discovered %s:%s', host, port)
                new.append(self.connect(host, port))
            elif not conn.alive():
                logger.info('Reconnecting to %s:%s', host, port)
                if conn.connect():
                    conn.setblocking(0)
                    self.reconnected(conn)
            else:
                logger.debug('Connection to %s:%s still alive', host, port)

        # And return all the new connections
        return [conn for conn in new if conn]

    def check_connections(self):
        '''Connect to all the appropriate instances'''
        logger.info('Checking connections')
        if self._lookupd:
            self.discover(self._topic)

        # Make sure we're connected to all the prescribed hosts
        for hostspec in self._nsqd_tcp_addresses:
            logger.debug('Checking nsqd instance %s', hostspec)
            host, port = hostspec.split(':')
            port = int(port)
            conn = self._connections.get((host, port), None)
            # If there is no connection to it, we have to try to connect
            if not conn:
                logger.info('Connecting to %s:%s', host, port)
                self.connect(host, port)
            elif not conn.alive():
                # If we've connected to it before, but it's no longer alive,
                # we'll have to make a decision about when to try to reconnect
                # to it, if we need to reconnect to it at all
                if conn.ready_to_reconnect():
                    logger.info('Reconnecting to %s:%s', host, port)
                    if conn.connect():
                        conn.setblocking(0)
                        self.reconnected(conn)
            else:
                logger.debug('Checking freshness')
                now = time.time()
                time_check = math.ceil(now - self.last_recv_timestamp)
                if time_check >= ((self.heartbeat_interval * 2) / 1000.0):
                    if conn.ready_to_reconnect():
                        logger.info('Reconnecting to %s:%s', host, port)
                        if conn.connect():
                            conn.setblocking(0)
                            self.reconnected(conn)

    @contextmanager
    def connection_checker(self):
        '''Run periodic reconnection checks'''
        thread = ConnectionChecker(self)
        logger.info('Starting connection-checker thread')
        thread.start()
        try:
            yield thread
        finally:
            logger.info('Stopping connection-checker')
            thread.stop()
            logger.info('Joining connection-checker')
            thread.join()

    def connect(self, host, port):
        '''Connect to the provided host, port'''
        conn = connection.Connection(host, port,
            reconnection_backoff=self._reconnection_backoff,
            auth_secret=self._auth_secret,
            timeout=self._connect_timeout,
            **self._identify_options)
        if conn.alive():
            conn.setblocking(0)
        self.add(conn)
        return conn

    def reconnected(self, conn):
        '''Hook into when a connection has been reestablished'''

    def connections(self):
        '''Safely return a list of all our connections'''
        with self._lock:
            return self._connections.values()

    def added(self, conn):
        '''Hook into when a connection has been added'''

    def add(self, connection):
        '''Add a connection'''
        key = (connection.host, connection.port)
        with self._lock:
            if key not in self._connections:
                self._connections[key] = connection
                self.added(connection)
                return connection
            else:
                return None

    def remove(self, connection):
        '''Remove a connection'''
        key = (connection.host, connection.port)
        with self._lock:
            found = self._connections.pop(key, None)
        try:
            self.close_connection(found)
        except Exception as exc:
            logger.warn('Failed to close %s: %s', connection, exc)
        return found

    def close_connection(self, connection):
        '''A hook for subclasses when connections are closed'''
        connection.close()

    def close(self):
        '''Close this client down'''
        map(self.remove, self.connections())

    def read(self):
        '''Read from any of the connections that need it'''
        # We'll check all living connections
        connections = [c for c in self.connections() if c.alive()]

        if not connections:
            # If there are no connections, obviously we return no messages, but
            # we should wait the duration of the timeout
            time.sleep(self._timeout)
            return []

        # Not all connections need to be written to, so we'll only concern
        # ourselves with those that require writes
        writes = [c for c in connections if c.pending()]
        try:
            readable, writable, exceptable = select.select(
                connections, writes, connections, self._timeout)
        except exceptions.ConnectionClosedException:
            logger.exception('Tried selecting on closed client')
            return []
        except select.error:
            logger.exception('Error running select')
            return []

        # If we returned because the timeout interval passed, log it and return
        if not (readable or writable or exceptable):
            logger.debug('Timed out...')
            return []

        responses = []
        # For each readable socket, we'll try to read some responses
        for conn in readable:
            try:
                for res in conn.read():
                    # We'll capture heartbeats and respond to them automatically
                    if (isinstance(res, Response) and res.data == HEARTBEAT):
                        logger.info('Sending heartbeat to %s', conn)
                        conn.nop()
                        logger.debug('Setting last_recv_timestamp')
                        self.last_recv_timestamp = time.time()
                        continue
                    elif isinstance(res, Error):
                        nonfatal = (
                            exceptions.FinFailedException,
                            exceptions.ReqFailedException,
                            exceptions.TouchFailedException
                        )
                        if not isinstance(res.exception(), nonfatal):
                            # If it's not any of the non-fatal exceptions, then
                            # we have to close this connection
                            logger.error(
                                'Closing %s: %s', conn, res.exception())
                            self.close_connection(conn)
                    responses.append(res)
                    logger.debug('Setting last_recv_timestamp')
                    self.last_recv_timestamp = time.time()
            except exceptions.NSQException:
                logger.exception('Failed to read from %s', conn)
                self.close_connection(conn)
            except socket.error:
                logger.exception('Failed to read from %s', conn)
                self.close_connection(conn)

        # For each writable socket, flush some data out
        for conn in writable:
            try:
                conn.flush()
            except socket.error:
                logger.exception('Failed to flush %s', conn)
                self.close_connection(conn)

        # For each connection with an exception, try to close it and remove it
        # from our connections
        for conn in exceptable:
            self.close_connection(conn)

        return responses

    @contextmanager
    def random_connection(self):
        '''Pick a random living connection'''
        # While at the moment there's no need for this to be a context manager
        # per se, I would like to use that interface since I anticipate
        # adding some wrapping around it at some point.
        yield random.choice(
            [conn for conn in self.connections() if conn.alive()])

    def wait_response(self):
        '''Wait for a response'''
        responses = self.read()
        while not responses:
            responses = self.read()
        return responses

    def wait_write(self, client):
        '''Wait until the specific client has written the message'''
        while client.pending():
            self.read()

    def pub(self, topic, message):
        '''Publish the provided message to the provided topic'''
        with self.random_connection() as client:
            client.pub(topic, message)
            return self.wait_response()

    def mpub(self, topic, *messages):
        '''Publish messages to a topic'''
        with self.random_connection() as client:
            client.mpub(topic, *messages)
            return self.wait_response()



'''Our clients for interacting with various clients'''

from decorator import decorator
import requests
import url

from .. import json, logger
from ..exceptions import NSQException


@decorator
def wrap(function, *args, **kwargs):
    '''Wrap a function that returns a request with some exception handling'''
    try:
        req = function(*args, **kwargs)
        logger.debug('Got %s: %s', req.status_code, req.content)
        if req.status_code == 200:
            return req
        else:
            raise ClientException(req.reason, req.content)
    except ClientException:
        raise
    except Exception as exc:
        raise ClientException(exc)


@decorator
def json_wrap(function, *args, **kwargs):
    '''Return the json content of a function that returns a request'''
    try:
        # Some responses have data = None, but they generally signal a
        # successful API call as well.
        response = json.loads(function(*args, **kwargs).content)
        if 'data' in response:
            return response['data'] or True
        else:
            return response
    except Exception as exc:
        raise ClientException(exc)


@decorator
def ok_check(function, *args, **kwargs):
    '''Ensure that the response body is OK'''
    req = function(*args, **kwargs)
    if req.content.lower() != 'ok':
        raise ClientException(req.content)
    return req.content


class ClientException(NSQException):
    '''An exception class for all client errors'''


class BaseClient(object):
    '''Base client class'''
    def __init__(self, target, **params):
        if isinstance(target, basestring):
            self._host = url.parse(target)
        elif isinstance(target, (tuple, list)):
            self._host = url.parse('http://%s:%s/' % target)
        else:
            raise TypeError('Host must be a string or tuple')
        self._params = params

    @wrap
    def get(self, path, *args, **kwargs):
        '''GET the provided endpoint'''
        target = self._host.relative(path).utf8
        if not isinstance(target, basestring):
            # on older versions of the `url` library, .utf8 is a method, not a property
            target = target()
        params = kwargs.get('params', {})
        params.update(self._params)
        kwargs['params'] = params
        logger.debug('GET %s with %s, %s', target, args, kwargs)
        return requests.get(target, *args, **kwargs)

    @wrap
    def post(self, path, *args, **kwargs):
        '''POST to the provided endpoint'''
        target = self._host.relative(path).utf8
        if not isinstance(target, basestring):
            # on older versions of the `url` library, .utf8 is a method, not a property
            target = target()
        params = kwargs.get('params', {})
        params.update(self._params)
        kwargs['params'] = params
        logger.debug('POST %s with %s, %s', target, args, kwargs)
        return requests.post(target, *args, **kwargs)



'''A Gevent-based client'''

from __future__ import absolute_import
import gevent.monkey
gevent.monkey.patch_all()

# The gevent client is actually the same as the synchronous client, just with
# socket, thread, ssl, select and time patched. By importing this here, it makes
# a gevent-compatible client available from nsq.gevent.Client
from .client import Client
from .reader import Reader



'''Sockets that wrap different connection types'''

# Not all platforms support all types of sockets provided here. For those that
# are not available, the corresponding socket wrapper is imported as None.

from .. import logger

# Snappy support
try:
    from .snappy import SnappySocket
except ImportError:  # pragma: no cover
    logger.debug('Snappy compression not supported')
    SnappySocket = None


# Deflate support
try:
    from .deflate import DeflateSocket
except ImportError:  # pragma: no cover
    logger.debug('Deflate compression not supported')
    DeflateSocket = None


# The TLS socket
try:
    from .tls import TLSSocket
except ImportError:  # pragma: no cover
    logger.warn('TLS not supported')
    TLSSocket = None



import mock

from nsq import client
from nsq import response
from nsq import constants
from nsq import exceptions
from nsq.http import ClientException

from common import HttpClientIntegrationTest, MockedConnectionTest
from contextlib import contextmanager
import errno
import select
import socket


class TestClientNsqd(HttpClientIntegrationTest):
    '''Test our client class'''
    nsqd_ports = (14150,)

    def setUp(self):
        '''Return a new client'''
        HttpClientIntegrationTest.setUp(self)
        hosts = ['localhost:%s' % port for port in self.nsqd_ports]
        self.client = client.Client(nsqd_tcp_addresses=hosts)

    def test_connect_nsqd(self):
        '''Can successfully establish connections'''
        connections = self.client.connections()
        self.assertEqual(len(connections), 1)
        for connection in connections:
            self.assertTrue(connection.alive())

    def test_add_added(self):
        '''Connect invokes self.connected'''
        connection = mock.Mock()
        with mock.patch.object(self.client, 'added'):
            self.client.add(connection)
            self.client.added.assert_called_with(connection)

    def test_add_existing(self):
        '''Adding an existing connection returns None'''
        connection = self.client.connections()[0]
        self.assertEqual(self.client.add(connection), None)

    def test_remove_exception(self):
        '''If closing a connection raises an exception, remove still works'''
        connection = self.client.connections()[0]
        with mock.patch.object(connection, 'close', side_effect=Exception):
            self.assertEqual(self.client.remove(connection), connection)

    def test_honors_identify_options(self):
        '''Sends along identify options to each connection as it's created'''
        with mock.patch('nsq.client.connection.Connection') as MockConnection:
            with mock.patch.object(
                self.client, '_identify_options', {'foo': 'bar'}):
                self.client.connect('foo', 'bar')
                MockConnection.assert_called_with('foo', 'bar',
                    reconnection_backoff=None, auth_secret=None, foo='bar', timeout=None)

    def test_conection_checker(self):
        '''Spawns and starts a connection checker'''
        with self.client.connection_checker() as checker:
            self.assertTrue(checker.is_alive())
        self.assertFalse(checker.is_alive())

    def test_read_closed(self):
        '''Recovers from reading on a closed connection'''
        conn = self.client.connections()[0]
        with mock.patch.object(conn, 'alive', return_value=True):
            with mock.patch.object(conn, '_socket', None):
                # This test passes if no exception in raised
                self.client.read()

    def test_read_select_err(self):
        '''Recovers from select errors'''
        with mock.patch('nsq.client.select.select') as mock_select:
            mock_select.side_effect = select.error(errno.EBADF)
            # This test passes if no exception is raised
            self.client.read()


class TestClientLookupd(HttpClientIntegrationTest):
    '''Test our client class'''
    def setUp(self):
        '''Return a new client'''
        HttpClientIntegrationTest.setUp(self)
        self.client = client.Client(topic=self.topic, lookupd_http_addresses=['http://localhost:14161'])

    def test_connected(self):
        '''Can successfully establish connections'''
        connections = self.client.connections()
        self.assertEqual(len(connections), 1)
        for connection in connections:
            self.assertTrue(connection.alive())

    def test_asserts_topic(self):
        '''If nsqlookupd servers are provided, asserts a topic'''
        self.assertRaises(
            AssertionError, client.Client, lookupd_http_addresses=['foo'])

    def test_client_exception(self):
        '''Is OK when discovery fails'''
        with mock.patch('nsq.client.nsqlookupd.Client') as MockClass:
            instance = MockClass.return_value
            instance.lookup.side_effect = ClientException
            self.client = client.Client(
                lookupd_http_addresses=['http://localhost:1234'],
                topic='foo')

    def test_discover_connected(self):
        '''Doesn't freak out when rediscovering established connections'''
        before = self.client.connections()
        self.client.discover(self.topic)
        self.assertEqual(self.client.connections(), before)

    def test_discover_closed(self):
        '''Reconnects to discovered servers that have closed connections'''
        for conn in self.client.connections():
            conn.close()
        state = [conn.alive() for conn in self.client.connections()]
        self.assertEqual(state, [False])
        self.client.discover(self.topic)
        state = [conn.alive() for conn in self.client.connections()]
        self.assertEqual(state, [True])

    def test_auth_secret(self):
        '''If an auth secret is provided, it passes it to nsqlookupd'''
        with mock.patch('nsq.client.nsqlookupd.Client') as MockClient:
            client.Client(topic=self.topic, lookupd_http_addresses=['foo'], auth_secret='hello')
            MockClient.assert_called_with('foo', access_token='hello')


class TestClientMultiple(MockedConnectionTest):
    '''Tests for our client class'''
    @contextmanager
    def readable(self, connections):
        '''With all the connections readable'''
        value = (connections, [], [])
        with mock.patch('nsq.client.select.select', return_value=value):
            yield

    @contextmanager
    def writable(self, connections):
        '''With all the connections writable'''
        value = ([], connections, [])
        with mock.patch('nsq.client.select.select', return_value=value):
            yield

    @contextmanager
    def exceptable(self, connections):
        '''With all the connections exceptable'''
        value = ([], [], connections)
        with mock.patch('nsq.client.select.select', return_value=value):
            yield

    def test_multi_read(self):
        '''Can read from multiple sockets'''
        # With all the connections read-ready
        for conn in self.connections:
            conn.response('hello')
        with self.readable(self.connections):
            found = self.client.read()
            self.assertEqual(len(found), 2)
            for res in found:
                self.assertIsInstance(res, response.Response)
                self.assertEqual(res.data, 'hello')

    def test_heartbeat(self):
        '''Sends a nop on connections that have received a heartbeat'''
        for conn in self.connections:
            conn.response(constants.HEARTBEAT)
        with self.readable(self.connections):
            self.assertEqual(self.client.read(), [])
            for conn in self.connections:
                conn.nop.assert_called_with()

    def test_closes_on_fatal(self):
        '''All but a few errors are considered fatal'''
        self.connections[0].error(exceptions.InvalidException)
        with self.readable(self.connections):
            self.client.read()
            self.assertFalse(self.connections[0].alive())

    def test_nonfatal(self):
        '''Nonfatal errors keep the connection open'''
        self.connections[0].error(exceptions.FinFailedException)
        with self.readable(self.connections):
            self.client.read()
            self.assertTrue(self.connections[0].alive())

    def test_passes_errors(self):
        '''The client's read method should now swallow Error responses'''
        self.connections[0].error(exceptions.InvalidException)
        with self.readable(self.connections):
            res = self.client.read()
            self.assertEqual(len(res), 1)
            self.assertIsInstance(res[0], response.Error)
            self.assertEqual(res[0].data, exceptions.InvalidException.name)

    def test_closes_on_exception(self):
        '''If a connection gets an exception, it closes it'''
        # Pick a connection to have throw an exception
        conn = self.connections[0]
        with mock.patch.object(
            conn, 'read', side_effect=exceptions.NSQException):
            with self.readable(self.connections):
                self.client.read()
                self.assertFalse(conn.alive())

    def test_closes_on_read_socket_error(self):
        '''If a connection gets a socket error, it closes it'''
        # Pick a connection to have throw an exception
        conn = self.connections[0]
        with mock.patch.object(
            conn, 'read', side_effect=socket.error):
            with self.readable(self.connections):
                self.client.read()
                self.assertFalse(conn.alive())

    def test_closes_on_flush_socket_error(self):
        '''If a connection fails to flush, it gets closed'''
        # Pick a connection to have throw an exception
        conn = self.connections[0]
        with mock.patch.object(
            conn, 'flush', side_effect=socket.error):
            with self.writable(self.connections):
                self.client.read()
                self.assertFalse(conn.alive())

    def test_read_writable(self):
        '''Read flushes any writable connections'''
        with self.writable(self.connections):
            self.client.read()
            for conn in self.connections:
                conn.flush.assert_called_with()

    def test_read_exceptions(self):
        '''Closes connections with socket errors'''
        with self.exceptable(self.connections):
            self.client.read()
            for conn in self.connections:
                self.assertFalse(conn.alive())

    def test_read_timeout(self):
        '''Logs a message when our read loop finds nothing because of timeout'''
        with self.readable([]):
            with mock.patch('nsq.client.logger') as MockLogger:
                self.client.read()
                MockLogger.debug.assert_called_with('Timed out...')

    def test_read_with_no_connections(self):
        '''Attempting to read with no connections'''
        with mock.patch.object(self.client, 'connections', return_value=[]):
            self.assertEqual(self.client.read(), [])

    def test_read_sleep_no_connections(self):
        '''Sleeps for timeout if no connections'''
        with mock.patch.object(self.client, '_timeout', 5):
            with mock.patch.object(self.client, 'connections', return_value=[]):
                with mock.patch('nsq.client.time.sleep') as mock_sleep:
                    self.client.read()
                    mock_sleep.assert_called_with(self.client._timeout)

    def test_random_connection(self):
        '''Yields a random client'''
        found = []
        for _ in xrange(20):
            with self.client.random_connection() as conn:
                found.append(conn)
        self.assertEqual(set(found), set(self.client.connections()))

    def test_wait_response(self):
        '''Waits until a response is available'''
        with mock.patch.object(
            self.client, 'read', side_effect=[[], ['hello']]):
            self.assertEqual(self.client.wait_response(), ['hello'])

    def test_wait_write(self):
        '''Waits until a command has been sent'''
        connection = mock.Mock()
        with mock.patch.object(self.client, 'read'):
            connection.pending = mock.Mock(side_effect=[True, False])
            self.client.wait_write(connection)
            self.assertTrue(connection.pending.called)

    def test_pub(self):
        '''Pub called on a random connection and waits for a response'''
        connection = mock.Mock()
        with mock.patch.object(
            self.client, 'connections', return_value=[connection]):
            with mock.patch.object(
                self.client, 'wait_response', return_value=['response']):
                self.assertEqual(self.client.pub('foo', 'bar'), ['response'])
                connection.pub.assert_called_with('foo', 'bar')

    def test_mpub(self):
        '''Mpub called on a random connection and waits for a response'''
        connection = mock.Mock()
        messages = ['hello', 'how', 'are', 'you']
        with mock.patch.object(
            self.client, 'connections', return_value=[connection]):
            with mock.patch.object(
                self.client, 'wait_response', return_value=['response']):
                self.assertEqual(
                    self.client.mpub('foo', *messages), ['response'])
                connection.mpub.assert_called_with('foo', *messages)

    def test_not_ready_to_reconnect(self):
        '''Does not try to reconnect connections that are not ready'''
        conn = self.connections[0]
        conn.close()
        conn.ready_to_reconnect.return_value = False
        self.client.check_connections()
        self.assertFalse(conn.connect.called)

    def test_ready_to_reconnect(self):
        '''Tries to reconnect when ready'''
        conn = self.connections[0]
        conn.close()
        conn.ready_to_reconnect.return_value = True
        self.client.check_connections()
        self.assertTrue(conn.connect.called)

    def test_set_blocking(self):
        '''Sets blocking to 0 when reconnecting'''
        conn = self.connections[0]
        conn.close()
        conn.ready_to_reconnect.return_value = True
        conn.connect.return_value = True
        self.client.check_connections()
        conn.setblocking.assert_called_with(0)

    def test_calls_reconnected(self):
        '''Sets blocking to 0 when reconnecting'''
        conn = self.connections[0]
        conn.close()
        conn.ready_to_reconnect.return_value = True
        conn.connect.return_value = True
        with mock.patch.object(self.client, 'reconnected'):
            self.client.check_connections()
            self.client.reconnected.assert_called_with(conn)


class TestClientNsqdWithConnectTimeout(HttpClientIntegrationTest):
    '''Test our client class when a connection timeout is set'''
    nsqd_ports = (14150,)

    def test_connect_timeout(self):
        HttpClientIntegrationTest.setUp(self)
        hosts = ['localhost:%s' % port for port in self.nsqd_ports]
        connect_timeout = 2.0
        self.client = client.Client(nsqd_tcp_addresses=hosts, connect_timeout=connect_timeout)
        self.assertEqual(self.client._connect_timeout, connect_timeout)



#! /usr/bin/env python

import sys
from nsq import __version__

extra = {}

from setuptools import setup
if sys.version_info >= (3,):
    extra['use_2to3'] = True


setup(name               = 'nsq-py',
    version              = __version__,
    description          = 'NSQ for Python With Pure Sockets',
    url                  = 'http://github.com/dlecocq/nsq-py',
    author               = 'Dan Lecocq',
    author_email         = 'dan@moz.com',
    license              = "MIT License",
    keywords             = 'nsq, queue',
    packages             = ['nsq', 'nsq.http', 'nsq.sockets'],
    package_dir          = {'nsq': 'nsq', 'nsq.http': 'nsq/http', 'nsq.sockets': 'nsq/sockets'},
    classifiers          = [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ],
    install_requires=[
        'requests',
        'decorator',
        'url',
        'statsd'
    ],
    tests_requires=[
        'nose',
        'coverage'
    ],
    scripts = [
        'bin/nsqlookupd-statsd'
    ],
    **extra
)



'''Base socket wrapper'''


class SocketWrapper(object):
    '''Wraps a socket in another layer'''
    # Methods for which we the default should be to simply pass through to the
    # underlying socket
    METHODS = (
        'accept', 'bind', 'close', 'connect', 'fileno', 'getpeername',
        'getsockname', 'getsockopt', 'setsockopt', 'gettimeout', 'settimeout',
        'setblocking', 'listen', 'makefile', 'shutdown'
    )

    @classmethod
    def wrap_socket(cls, socket, **options):
        '''Returns a socket-like object that transparently does compression'''
        return cls(socket, **options)

    def __init__(self, socket):
        self._socket = socket
        for method in self.METHODS:
            # Check to see if this class overrides this method, and if not, then
            # we should have it simply map through to the underlying socket
            if not hasattr(self, method):
                setattr(self, method, getattr(self._socket, method))

    def send(self, data, flags=0):
        '''Same as socket.send'''
        raise NotImplementedError()

    def sendall(self, data, flags=0):
        '''Same as socket.sendall'''
        count = len(data)
        while count:
            sent = self.send(data, flags)
            # This could probably be a buffer object
            data = data[sent:]
            count -= sent

    def recv(self, nbytes, flags=0):
        '''Same as socket.recv'''
        raise NotImplementedError()

    def recv_into(self, buff, nbytes, flags=0):
        '''Same as socket.recv_into'''
        raise NotImplementedError('Wrapped sockets do not implement recv_into')



import inspect
import struct

from .constants import FRAME_TYPE_RESPONSE, FRAME_TYPE_MESSAGE, FRAME_TYPE_ERROR
from . import exceptions

from contextlib import contextmanager
import socket
import sys


class Response(object):
    '''A response from NSQ'''
    FRAME_TYPE = FRAME_TYPE_RESPONSE

    __slots__ = ('connection', 'frame_type', 'data')

    @staticmethod
    def from_raw(conn, raw):
        '''Return a new response from a raw buffer'''
        frame_type = struct.unpack('>l', raw[0:4])[0]
        message = raw[4:]
        if frame_type == FRAME_TYPE_MESSAGE:
            return Message(conn, frame_type, message)
        elif frame_type == FRAME_TYPE_RESPONSE:
            return Response(conn, frame_type, message)
        elif frame_type == FRAME_TYPE_ERROR:
            return Error(conn, frame_type, message)
        else:
            raise TypeError('Unknown frame type: %s' % frame_type)

    @classmethod
    def pack(cls, data):
        '''Pack the provided data into a Response'''
        return struct.pack('>ll', len(data) + 4, cls.FRAME_TYPE) + data

    def __init__(self, conn, frame_type, data):
        self.connection = conn
        self.data = data
        self.frame_type = frame_type

    def __str__(self):
        return '%s - %s' % (self.__class__.__name__, self.data)

    def __eq__(self, other):
        return (
            (self.frame_type == other.frame_type) and
            (self.connection == other.connection) and
            (self.data == other.data))


class Message(Response):
    '''A message'''
    FRAME_TYPE = FRAME_TYPE_MESSAGE

    format = '>qH16s'
    size = struct.calcsize(format)

    __slots__ = ('timestamp', 'attempts', 'id', 'body', 'processed')

    @classmethod
    def pack(cls, timestamp, attempts, _id, data):
        return struct.pack(
            '>llqH16s',
            len(data) + cls.size + 4,
            cls.FRAME_TYPE,
            timestamp,
            attempts,
            _id) + data

    def __init__(self, conn, frame_type, data):
        Response.__init__(self, conn, frame_type, data)
        self.timestamp, self.attempts, self.id = struct.unpack(
            self.format, data[:self.size])
        self.body = data[self.size:]
        self.processed = False

    def __str__(self):
        return '%s - %i %i %s %s' % (
            self.__class__.__name__,
            self.timestamp,
            self.attempts,
            self.id,
            self.body)

    def fin(self):
        '''Indicate that this message is finished processing'''
        self.connection.fin(self.id)
        self.processed = True

    def req(self, timeout):
        '''Re-queue a message'''
        self.connection.req(self.id, timeout)
        self.processed = True

    def touch(self):
        '''Reset the timeout for an in-flight message'''
        self.connection.touch(self.id)

    def delay(self):
        '''How long to delay its requeueing'''
        return 60

    @contextmanager
    def handle(self):
        '''Make sure this message gets either 'fin' or 'req'd'''
        try:
            yield self
        except:
            # Requeue the message and raise the original exception
            typ, value, trace = sys.exc_info()
            if not self.processed:
                try:
                    self.req(self.delay())
                except socket.error:
                    self.connection.close()
            raise typ, value, trace
        else:
            if not self.processed:
                try:
                    self.fin()
                except socket.error:
                    self.connection.close()


class Error(Response):
    '''An error'''
    FRAME_TYPE = FRAME_TYPE_ERROR

    # A mapping of the response string to the appropriate exception
    mapping = {}

    @classmethod
    def find(cls, name):
        '''Find the exception class by name'''
        if not cls.mapping:  # pragma: no branch
            for _, obj in inspect.getmembers(exceptions):
                if inspect.isclass(obj):
                    if issubclass(obj, exceptions.NSQException):  # pragma: no branch
                        if hasattr(obj, 'name'):
                            cls.mapping[obj.name] = obj
        klass = cls.mapping.get(name)
        if klass == None:
            raise TypeError('No matching exception for %s' % name)
        return klass

    def exception(self):
        '''Return an instance of the corresponding exception'''
        code, _, message = self.data.partition(' ')
        return self.find(code)(message)



'''Exception classes'''


class NSQException(StandardError):
    '''Base class for all exceptions in this library'''


class ConnectionTimeoutException(NSQException):
    '''Connection instantiation timed out'''


class ConnectionClosedException(NSQException):
    '''Trying to use a closed connection as if it's alive'''


class UnsupportedException(NSQException):
    '''When a requested feature cannot be used'''


class TimeoutException(NSQException):
    '''Exception for failing a timeout'''


class InvalidException(NSQException):
    '''Exception for E_INVALID'''
    name = 'E_INVALID'


class BadBodyException(NSQException):
    '''Exception for E_BAD_BODY'''
    name = 'E_BAD_BODY'


class BadTopicException(NSQException):
    '''Exception for E_BAD_TOPIC'''
    name = 'E_BAD_TOPIC'


class BadChannelException(NSQException):
    '''Exception for E_BAD_CHANNEL'''
    name = 'E_BAD_CHANNEL'


class BadMessageException(NSQException):
    '''Exception for E_BAD_MESSAGE'''
    name = 'E_BAD_MESSAGE'


class PubFailedException(NSQException):
    '''Exception for E_PUB_FAILED'''
    name = 'E_PUB_FAILED'


class MpubFailedException(NSQException):
    '''Exception for E_MPUB_FAILED'''
    name = 'E_MPUB_FAILED'


class FinFailedException(NSQException):
    '''Exception for E_FIN_FAILED'''
    name = 'E_FIN_FAILED'


class ReqFailedException(NSQException):
    '''Exception for E_REQ_FAILED'''
    name = 'E_REQ_FAILED'


class TouchFailedException(NSQException):
    '''Exception for E_TOUCH_FAILED'''
    name = 'E_TOUCH_FAILED'
